local mainTank
local registeredHealers = {}
local unregisteredHealers = {}
local loathebMainFrame = CreateFrame("FRAME", "LoathebMainFrame", UIParent)
loathebMainFrame:SetPoint("CENTER", UIParent, "CENTER")
local COMM_PREFIX = "EZLoatheb"
C_ChatInfo.RegisterAddonMessagePrefix(COMM_PREFIX)

local allTimers = {
    ["UPDATE_MED"] = 10,
    ["UPDATE_SEC"] = 1,
    ["UPDATE_FREQUENT"] = 0.2,
}

local corruptedMindID = {
    [29184] = true, [29195] = true, [29197] = true, [29199] = true
}

-- TODO:
-- Mark unresponsive
-- Test in combat

-- Cache commonly used functions
local UnitHealthMax = UnitHealthMax;
local UnitHealth = UnitHealth;
local InCombatLockdown = InCombatLockdown;

local needsUpdate = false
local mainTankHealthDiff = 0
local myHealAmount = 2300

local CTL = _G.ChatThrottleLib
local function sendAddonMessage(msg, target)
    if( target and strlen(msg) <= 240 ) then
        if CTL then
			CTL:SendAddonMessage("BULK", COMM_PREFIX, msg, "WHISPER", target)
		end
	end
end

-- Check Timer for Frequent Updates
local function checkResetTimer(timerName, nextTick, resetTime)
	if allTimers[timerName] > 0 then
		allTimers[timerName] = allTimers[timerName] - nextTick
		if allTimers[timerName] <= 0 then
			allTimers[timerName] = resetTime;
			return true;
		end
	end
	return false;
end

local function calculateHealAmount()
    local classFilename = UnitClassBase("player")
    local spellHeal = GetSpellBonusHealing()

    if classFilename == "PALADIN" then
        return math.floor(math.min(1700 + 0.7 * spellHeal, 3200))
    elseif classFilename == "PRIEST" then
        return math.floor(math.min(2100 + 0.85 * spellHeal, 3200))
    elseif classFilename == "DRUID" then
        return math.floor(math.min(2000 + spellHeal, 3200))
    elseif classFilename == "SHAMAN" then
        return math.floor(math.min(1700 + 0.85 * spellHeal, 3200))
    else
        return 0
    end
end

local function getMainTank()
    local partySize = GetNumGroupMembers()
    for i=1, partySize do
        key = "raid" .. i
        if ((select(10, GetRaidRosterInfo(i))) == "MAINTANK") then
            mainTank = key
            return
        end
    end
    print("EZLoatheb: A single tank needs to be marked as Main Tank in /raidinfo")
end

local function updateHealerList()
    myHealAmount = calculateHealAmount()
    unregisteredHealers = {}
    local partySize = GetNumGroupMembers()
    for i=1, partySize do
        key = "raid" .. i
        local classFilename = UnitClassBase(key)
        if classFilename == "PALADIN" or classFilename == "PRIEST" or classFilename == "DRUID" or classFilename == "SHAMAN" then
            unregisteredHealers[#unregisteredHealers + 1] = {unitID = key, ezLoatheb = false, unitName = UnitName(key)}
        end
    end
    for i=1, #unregisteredHealers do
        if UnitIsConnected(unregisteredHealers[i].unitID) then
            sendAddonMessage("ping," .. myHealAmount, (UnitName(unregisteredHealers[i].unitID)))
        end
    end
end

local function checkPlayerDebuff(unitID, playerInfo)
    local i = 1
    local debbuffSlot = 1
    local name, icon, count, debuffType, duration, expirationTime, source, isStealable, nameplateShowPersonal, spellID, canApplyAura, isBossDebuff, castByPlayer, nameplateShowAll, timeMod = UnitDebuff(unitID, i)
    local corruptedMind = false
    while (name) do
        i = i+1
        if corruptedMindID[spellID] then
            corruptedMind = true
        end
        name, icon, count, debuffType, duration, expirationTime, source, isStealable, nameplateShowPersonal, spellID, canApplyAura, isBossDebuff, castByPlayer, nameplateShowAll, timeMod = UnitDebuff(unitID, i)
    end
    if corruptedMind and playerInfo.healingDebuff == 0 then
        playerInfo.healingDebuff = 61
    end
end

-- Should not be called in combat
local function updateRegisteredHealerList()
    if InCombatLockdown() then
        needsUpdate = true
        return
    end
    registeredHealers = {}
    for i=1, #unregisteredHealers do
        if UnitIsUnit(unregisteredHealers[i].unitID, "player") or unregisteredHealers[i].ezLoatheb then
            registeredHealers[#registeredHealers + 1] = {unitID = unregisteredHealers[i].unitID, unitName = unregisteredHealers[i].unitName, healAmount = unregisteredHealers[i].healAmount or 2300, healingDebuff = 0, active = true}
        end
    end
end

local function updateHealerRegistration(sender, senderHealAmount)
    if InCombatLockdown() then
        needsUpdate = true
        return false
    end
    for i=1, #unregisteredHealers do
        if string.find(sender, unregisteredHealers[i].unitName .. "-") then
            if not unregisteredHealers[i].ezLoatheb then
                unregisteredHealers[i].ezLoatheb = true
                unregisteredHealers[i].healAmount = senderHealAmount
                return true
            end
        end
    end
    return false
end

local function createUIFrames()
    if InCombatLockdown() then
        needsUpdate = true
        return
    end
    loathebMainFrame:SetMovable(true)
    loathebMainFrame:EnableMouse(true)
    loathebMainFrame:RegisterForDrag("LeftButton")
    loathebMainFrame:SetScript("OnDragStart", function()
        loathebMainFrame:StartMoving()
    end)
    loathebMainFrame:SetScript("OnDragStop", loathebMainFrame.StopMovingOrSizing)
    if not loathebMainFrame.texture then loathebMainFrame.texture = loathebMainFrame:CreateTexture(nil, "BACKGROUND"); end
    loathebMainFrame.texture:SetAllPoints(loathebMainFrame)
    loathebMainFrame.texture:SetColorTexture(0.3,0.3,0.3)
    loathebMainFrame.texture:SetAlpha(0.5)
    loathebMainFrame:SetSize(120, 15 + 15*#registeredHealers)

    -- Clickable Indicator
    if not loathebMainFrame.indicatorFrame then loathebMainFrame.indicatorFrame = CreateFrame("FRAME", "LoathebIndicatorFrame", loathebMainFrame) end;
    loathebMainFrame.indicatorFrame:SetMovable(true)
    loathebMainFrame.indicatorFrame:EnableMouse(true)
    loathebMainFrame.indicatorFrame:RegisterForDrag("LeftButton")
    loathebMainFrame.indicatorFrame:SetScript("OnDragStart", function()
        loathebMainFrame:StartMoving()
    end)
    loathebMainFrame.indicatorFrame:SetScript("OnDragStop", function() loathebMainFrame:StopMovingOrSizing() end)

    loathebMainFrame.indicatorFrame:SetSize(50, 50)
    loathebMainFrame.indicatorFrame:SetPoint("CENTER", loathebMainFrame, "CENTER", -100, 0)
    if not loathebMainFrame.indicatorFrame.texture then loathebMainFrame.indicatorFrame.texture = loathebMainFrame.indicatorFrame:CreateTexture(nil, "BACKGROUND"); end
    loathebMainFrame.indicatorFrame.texture:SetAllPoints(loathebMainFrame.indicatorFrame)
    loathebMainFrame.indicatorFrame.texture:SetColorTexture(0.3, 0.3, 0.3)
    loathebMainFrame.indicatorFrame.texture:SetAlpha(0.75)
    if not loathebMainFrame.indicatorFrame.text then loathebMainFrame.indicatorFrame.text = loathebMainFrame.indicatorFrame:CreateFontString(nil, "OVERLAY", "GameFontNormal"); end
    loathebMainFrame.indicatorFrame.text:SetPoint("CENTER", loathebMainFrame.indicatorFrame, "CENTER", 0, 0)
    loathebMainFrame.indicatorFrame.text:SetWidth(45)

    -- Display registered healers
    local j = 1
    for i=1, #registeredHealers do
        local textKey = "text" .. i
        if not loathebMainFrame[textKey] then loathebMainFrame[textKey] = loathebMainFrame:CreateFontString(nil, "OVERLAY", "GameFontNormal"); end
        loathebMainFrame[textKey]:Show()
        loathebMainFrame[textKey]:SetPoint("LEFT", loathebMainFrame, "TOPLEFT", 10, -15*i)
        loathebMainFrame[textKey]:SetTextColor(0, 0.8, 0)
        loathebMainFrame[textKey]:SetText(registeredHealers[i].unitName)
        j = j + 1

    end
    for k = j, 40 do
        local textKey = "text" .. k
        if loathebMainFrame[textKey] then
            loathebMainFrame[textKey]:Hide()
        else
            return
        end
    end
end


local function updateHealingRecommendation()
    healAmount = 0
    -- Get available healing amount from higher priority healers
    -- Check if player is CD first
    local i = 1
    while registeredHealers[i] and not UnitIsUnit(registeredHealers[i].unitID, "player") do 
        if registeredHealers[i].healingDebuff == 0 and registeredHealers[i].active then
            healAmount = healAmount + registeredHealers[i].healAmount
        end
        i = i + 1
    end
    if not registeredHealers[i] then
        return
    end
    if registeredHealers[i].healingDebuff > 0 then
        -- CD
        loathebMainFrame.indicatorFrame.texture:SetColorTexture(0.3, 0.3, 0.3)
        loathebMainFrame.indicatorFrame.text:SetText(registeredHealers[i].healingDebuff)
    elseif mainTankHealthDiff > healAmount + myHealAmount then
        -- GREEN: Cast Heal
        loathebMainFrame.indicatorFrame.texture:SetColorTexture(0, 1, 0)
        loathebMainFrame.indicatorFrame.text:SetText("HEAL")
    elseif healAmount < 4000 then
        -- YELLOW: Get Ready To Heal
        loathebMainFrame.indicatorFrame.texture:SetColorTexture(0.8, 0.8, 0.2)
        loathebMainFrame.indicatorFrame.text:SetText("GET READY")
    else
        -- Not next in line to heal
        loathebMainFrame.indicatorFrame.texture:SetColorTexture(0.3, 0.3, 0.3)
        loathebMainFrame.indicatorFrame.text:SetText("WAIT")
    end
end

local function updateHealerDisplay()
    for i=1, #registeredHealers do
        local textKey = "text" .. i
        loathebMainFrame[textKey]:SetText(registeredHealers[i].unitName .. " " .. registeredHealers[i].healingDebuff)
        if registeredHealers[i].healingDebuff > 0 then
            loathebMainFrame[textKey]:SetTextColor(0.8, 0, 0)
        else
            loathebMainFrame[textKey]:SetTextColor(0, 0.8, 0)
        end
    end
end

local function healerDebuffTick()
    for i=1, #registeredHealers do
        if registeredHealers[i].healingDebuff > 0 then
            registeredHealers[i].healingDebuff = registeredHealers[i].healingDebuff - 1
        end
    end
end

local function UpdateFrequent(self, timeDelta)
    if checkResetTimer("UPDATE_FREQUENT", timeDelta, 0.2) then
        for i=1, #registeredHealers do
            checkPlayerDebuff(registeredHealers[i].unitID, registeredHealers[i])
        end
    end
    if checkResetTimer("UPDATE_SEC", timeDelta, 1) then
        healerDebuffTick()
        updateHealerDisplay()
        updateHealingRecommendation()
        if not InCombatLockdown() and needsUpdate then
            needsUpdate = false
            updateHealerList()
            createUIFrames()
        end
    end
    if checkResetTimer("UPDATE_MED", timeDelta, 10) then
        for i=1, #registeredHealers do
            --print(registeredHealers[i].unitID)
            --print(registeredHealers[i].healAmount)
        end
    end
end

-- This function for testing only
--[[
local function checkRegisteredHealerHealed(healerGUID)
    for i = 1, #registeredHealers do
        if UnitGUID(registeredHealers[i].unitID) == healerGUID then
            if registeredHealers[i].healingDebuff == 0 then
                registeredHealers[i].healingDebuff = 61
            end
        end
    end
end
]]

local function main()
    local eventHandlers = {}

    function eventHandlers:ADDON_LOADED(addonName)
        if addonName == "EZLoatheb" then
            getMainTank()
            updateHealerList()
            createUIFrames()
        end
    end

    -- Handler for party changes
    function eventHandlers:GROUP_ROSTER_UPDATE()
        getMainTank()
        updateHealerList()
        createUIFrames()
    end

    function eventHandlers:CHAT_MSG_ADDON(prefix, message, type, sender)
        if prefix == COMM_PREFIX then
            local messagePrefix, senderHealAmount = message:match("([^,]+),([^,]+)")
            if messagePrefix == "ping" then
                sendAddonMessage("pong," .. myHealAmount, sender);
                local updated = updateHealerRegistration(sender, tonumber(senderHealAmount))
                if updated then
                    updateRegisteredHealerList()
                    createUIFrames()
                end
            elseif messagePrefix == "pong" then
                updateHealerRegistration(sender, tonumber(senderHealAmount))
                updateRegisteredHealerList()
                createUIFrames()
            end
        end
    end

    function eventHandlers:UNIT_HEALTH(unitID)
        if mainTank and unitID == mainTank then
            local oldHealthVal = mainTankHealthDiff
            mainTankHealthDiff = UnitHealthMax(mainTank) - UnitHealth(mainTank)
            if oldHealthVal ~= mainTankHealthDiff then
                updateHealingRecommendation()
            end
        end
    end

    -- This is for testing only
    --[[
    function eventHandlers:COMBAT_LOG_EVENT_UNFILTERED()
        local timestamp, subevent, _, sourceGUID, sourceName, _, _, destGUID, destName, _, _, _, spellName, _, healAmount, oh, _, crit = CombatLogGetCurrentEventInfo()
        if subevent == "SPELL_HEAL" then
            checkRegisteredHealerHealed(sourceGUID)
        end
    end
    ]]

    -- Attach all handlers to discoMainFrame
    loathebMainFrame:SetScript("OnEvent", function(self, event, ...)
        eventHandlers[event](self, ...); -- call one of the functions above
    end)
    for k, v in pairs(eventHandlers) do
        loathebMainFrame:RegisterEvent(k); -- Register all events for which handlers have been defined
    end
    loathebMainFrame:SetScript("OnUpdate", UpdateFrequent)


end
local classFilename = UnitClassBase("player")
if classFilename == "PALADIN" or classFilename == "PRIEST" or classFilename == "DRUID" or classFilename == "SHAMAN" then
    main()
end